//+------------------------------------------------------------------+
//|                                                 TrailingLFSR.mqh |
//|                             Copyright 2000-2023, MetaQuotes Ltd. |
//|                                             https://www.mql5.com |
//+------------------------------------------------------------------+
#include <Expert\ExpertTrailing.mqh>
// wizard description start
//+----------------------------------------------------------------------+
//| Description of the class                                             |
//| Title=Trailing Stop based on number wall                             |
//| Type=Trailing                                                        |
//| Name=LFSR                                                            |
//| Class=CTrailingLFSR                                                  |
//| Page=                                                                |
//| Parameter=StopLevel,int,30,Stop Loss trailing level (in points)      |
//| Parameter=ProfitLevel,int,50,Take Profit trailing level (in points)  |
//| Parameter=Length,int,5,Length                                        |
//| Parameter=Solve,double,20.0,Solve Threshold                          |
//+----------------------------------------------------------------------+
// wizard description end
//+------------------------------------------------------------------+
//| Class CTrailingLFSR.                                             |
//| Purpose: Class of trailing stops with number wall.               |
//|              Derives from class CExpertTrailing.                 |
//+------------------------------------------------------------------+
class CTrailingLFSR : public CExpertTrailing
{
protected:
   //--- input parameters
   int               m_stop_level;
   int               m_profit_level;

   int               m_length;
   double            m_solve;

   CiMA              m_ma;

public:
                     CTrailingLFSR(void);
                    ~CTrailingLFSR(void);
   //--- methods of initialization of protected data
   void              StopLevel(int stop_level)
   {  m_stop_level = stop_level;
   }
   void              ProfitLevel(int profit_level)
   {  m_profit_level = profit_level;
   }
   virtual bool      ValidationSettings(void);
   //--- method of creating the indicator and timeseries
   virtual bool      InitIndicators(CIndicators *indicators);
   //---
   virtual bool      CheckTrailingStopLong(CPositionInfo *position, double &sl, double &tp);
   virtual bool      CheckTrailingStopShort(CPositionInfo *position, double &sl, double &tp);

   void              Length(int value)
   {  m_length = value;
   };

   void              Solve(double value)
   {  m_solve = value;
   };

   double            GetOutput(int Index, double &Solution);
   double            SetOutput(vector &True, vector &Predicted);

   vector            Get(vector &Top, vector &Left, vector &Right, vector &Center);
   vector            Set();

   double            Solvability(vector &V);

   struct            Srows
   {  vector         column[];
                     Srows() {};
                    ~Srows() {};
   };
   struct            Smatrix
   {  Srows          row[];
                     Smatrix() {};
                    ~Smatrix() {};
   };

   Smatrix           m_numberwall;
   //matrix<vector>  m_numberwall;
};
//+------------------------------------------------------------------+
//| Constructor                                                      |
//+------------------------------------------------------------------+
void CTrailingLFSR::CTrailingLFSR(void) : m_stop_level(30),
                                          m_profit_level(50), 
                                          m_length(5),
                                          m_solve(20.0)
{
//--- initialization of protected data
   m_used_series=USE_SERIES_CLOSE;
}
//+------------------------------------------------------------------+
//| Destructor                                                       |
//+------------------------------------------------------------------+
CTrailingLFSR::~CTrailingLFSR(void)
{
}
//+------------------------------------------------------------------+
//| Validation settings protected data.                              |
//+------------------------------------------------------------------+
bool CTrailingLFSR::ValidationSettings(void)
{  if(!CExpertTrailing::ValidationSettings())
      return(false);
//--- initial data checks
   if(m_profit_level != 0 && m_profit_level * (m_adjusted_point / m_symbol.Point()) < m_symbol.StopsLevel())
   {  printf(__FUNCTION__ + ": trailing Profit Level must be 0 or greater than %d", m_symbol.StopsLevel());
      return(false);
   }
   if(m_stop_level != 0 && m_stop_level * (m_adjusted_point / m_symbol.Point()) < m_symbol.StopsLevel())
   {  printf(__FUNCTION__ + ": trailing Stop Level must be 0 or greater than %d", m_symbol.StopsLevel());
      return(false);
   }
   ArrayResize(m_numberwall.row, 2 + m_length);
   for(int i = 0; i < 2 + m_length; i++)
   {  ArrayResize(m_numberwall.row[i].column, 2 + m_length);
   }
   vector _v0, _v1;
   _v0.Init(1);
   _v0.Fill(0.0);
   _v1.Init(1);
   _v1.Fill(1.0);
   for(int i = 0; i < 2 + m_length; i++)
   {  for(int ii = 0; ii < 2 + m_length; ii++)
      {  if((i == 1 && ii != 0) || (i != 0 && ii == 1))
         {  m_numberwall.row[i].column[ii] = _v1;
         }
         else
         {  m_numberwall.row[i].column[ii] = _v0;
         }
      }
   }
//--- ok
   return(true);
}
//+------------------------------------------------------------------+
//| Create indicators.                                               |
//+------------------------------------------------------------------+
bool CTrailingLFSR::InitIndicators(CIndicators *indicators)
{
//--- check pointer
   if(indicators == NULL)
      return(false);
//--- initialization of indicators and timeseries of additional filters
   if(!CExpertTrailing::InitIndicators(indicators))
      return(false);
   if(!m_ma.Create(m_symbol.Name(), m_period, m_length, 0, MODE_SMA, PRICE_CLOSE))
   {  return(false);
   }
//--- ok
   return(true);
}
//+------------------------------------------------------------------+
//| Checking trailing stop and/or profit for long position.          |
//+------------------------------------------------------------------+
bool CTrailingLFSR::CheckTrailingStopLong(CPositionInfo *position, double &sl, double &tp)
{
//--- check
   if(position == NULL)
      return(false);
   if(m_stop_level == 0)
      return(false);
//---
   double delta;
   double pos_sl = position.StopLoss();
   double base  = (pos_sl == 0.0) ? position.PriceOpen() : pos_sl;
   double price = m_symbol.Bid();
//---
   vector _t, _p;
   _p.Init(2);
   _t.CopyIndicatorBuffer(m_ma.Handle(), 0, 0, 2);
   double _s = 0.0;
   for(int i = 1; i >= 0; i--)
   {  _s = 0.0;
      _p[i] = GetOutput(i, _s);
   }
   double _o = SetOutput(_t, _p);
//---
   sl = EMPTY_VALUE;
   tp = EMPTY_VALUE;
   delta = m_stop_level * m_adjusted_point;
   if(_s < m_solve && _o > _t[0])
   {  delta = (_o-_t[0]);
   }
//---
   if(price - base > delta)
   {  sl = price - delta;
      if(m_profit_level != 0)
         tp = price + m_profit_level * m_adjusted_point;
   }
//---
   return(sl != EMPTY_VALUE);
}
//+------------------------------------------------------------------+
//| Checking trailing stop and/or profit for short position.         |
//+------------------------------------------------------------------+
bool CTrailingLFSR::CheckTrailingStopShort(CPositionInfo *position, double &sl, double &tp)
{
//--- check
   if(position == NULL)
      return(false);
   if(m_stop_level == 0)
      return(false);
//---
   double delta;
   double pos_sl = position.StopLoss();
   double base  = (pos_sl == 0.0) ? position.PriceOpen() : pos_sl;
   double price = m_symbol.Ask();
//---
   vector _t, _p;
   _p.Init(2);
   _t.CopyIndicatorBuffer(m_ma.Handle(), 0, 0, 2);
   double _s = 0.0;
   for(int i = 1; i >= 0; i--)
   {  _s = 0.0;
      _p[i] = GetOutput(i, _s);
   }
   double _o = SetOutput(_t, _p);
//---
   sl = EMPTY_VALUE;
   tp = EMPTY_VALUE;
   delta = m_stop_level * m_adjusted_point;
   if(_s < m_solve && _o < _t[0])
   {  delta = (_t[0] - _o);
   }
//---
   if(base - price > delta)
   {  sl = price + delta;
      if(m_profit_level != 0)
         tp = price - m_profit_level * m_adjusted_point;
   }
//---
   return(sl != EMPTY_VALUE);
}
//+------------------------------------------------------------------+
//| LFSR Output.                                                     |
//+------------------------------------------------------------------+
double CTrailingLFSR::GetOutput(int Index, double &Solution)
{  double _output = 0.0;
   vector _v;
   _v.CopyIndicatorBuffer(m_ma.Handle(), 0, Index, m_length);
   Solution = Solvability(_v);
   _v.Resize(m_length + 1);
   for(int i = 2; i < 2 + m_length; i++)
   {  for(int ii = 2; ii < 2 + m_length; ii++)
      {  if(i == 2)
         {  vector _vi;
            _vi.Init(1);
            _vi[0] = _v[m_length - ii + 1];
            m_numberwall.row[i].column[ii] = _vi;
         }
         else if(i == 3)
         {  vector _vi;
            _vi.Init(2);
            _vi[0] = m_numberwall.row[i - 1].column[ii][0];
            _vi[1] = -1.0 * m_numberwall.row[i - 1].column[ii - 1][0];
            m_numberwall.row[i].column[ii] = _vi;
         }
         else if(ii < m_length + 1)
         {  m_numberwall.row[i].column[ii] = Get(m_numberwall.row[i - 2].column[ii], m_numberwall.row[i - 1].column[ii - 1], m_numberwall.row[i - 1].column[ii + 1], m_numberwall.row[i - 1].column[ii]);
         }
      }
   }
   vector _u = Set();
   vector _x;
   _x.CopyIndicatorBuffer(m_ma.Handle(), 0, Index, m_length);
   _u.Resize(fmax(_u.Size(),_x.Size()));
   _x.Resize(fmax(_u.Size(),_x.Size()));
   vector _y = _u * _x;
   _output = _y.Sum();
   return(_output);
}
//+------------------------------------------------------------------+
//| Normalising Output to match Indicator Value                      |
//+------------------------------------------------------------------+
double CTrailingLFSR::SetOutput(vector &True, vector &Predicted)
{  return(True[1] - ((True[0] - True[1]) * ((Predicted[0] - Predicted[1]) / fmax(m_symbol.Point(), fabs(Predicted[0]) + fabs(Predicted[1])))));
}
//+------------------------------------------------------------------+
//| Get known Value                                                  |
//+------------------------------------------------------------------+
vector CTrailingLFSR::Get(vector &Top, vector &Left, vector &Right, vector &Center)
{  vector _cc, _lr, _cc_lr, _i_top;
   _cc = Center.Correlate(Center, VECTOR_CONVOLVE_FULL);
   _lr = Left.Correlate(Right, VECTOR_CONVOLVE_FULL);
   ulong _size = fmax(_cc.Size(), _lr.Size());
   _cc_lr.Init(_size);
   _cc.Resize(_size);
   _lr.Resize(_size);
   _cc_lr = _cc - _lr;
   _i_top = 1.0 / Top;
   vector _bottom = _cc_lr.Correlate(_i_top, VECTOR_CONVOLVE_FULL);
   return(_bottom);
}
//+------------------------------------------------------------------+
//| Set Unknown Value                                                |
//+------------------------------------------------------------------+
vector CTrailingLFSR::Set()
{  vector _formula = m_numberwall.row[m_length + 1].column[m_length + 1];
   vector _right;
   _right.Copy(_formula);
   _right.Resize(ulong(fmax(_formula.Size() - 1, 1.0)));
   double _solver = -1.0 * _formula[int(_formula.Size() - 1)];
   if(_solver != 0.0)
   {  _right /= _solver;
   }
   return(_right);
}
//+------------------------------------------------------------------+
//| Get Solvability of Shift Inputed Vector (aka Teoplitz Matrix).   |
//+------------------------------------------------------------------+
double CTrailingLFSR::Solvability(vector &V)
{  V /= m_symbol.Point();
   matrix _m;
   _m.Init(V.Size(), V.Size());
   for(ulong i = 0; i < V.Size(); i++)
   {  for(ulong ii = 0; ii < V.Size(); ii++)
      {  _m[i][ii] = V[(i + ii) % V.Size()];
      }
   }
   return(fabs(_m.Det() * pow(m_symbol.Point(), 3.0)));
}
//+------------------------------------------------------------------+
